import pytest
import os
from os.path import join
from cdm import get_session
from cdm.context import Context
from cdm.installer import Installer

def get_context(path):
    cache_dir = join(os.getcwd(), ".cdmcache")

    try:
        os.mkdir(cache_dir)
    except Exception as e:
        # TODO don't be so silly
        print e

    c = Context(root=path,
                dataset="test",
                session=get_session("test", True),
                cache_dir=cache_dir)
    return c


@pytest.fixture(scope="session")
def context():
    """
    returns a context with a whole bunch of stuff set up
    :return:
    """

    def fin():
        print "Teardown context"

    return get_context(os.getcwd())


def get_sample_context():
    sample = os.path.join(os.getcwd(), "tests/sample")
    return get_context(sample)


def get_installer(name):
    sample = os.path.join(os.getcwd(), "tests/{}".format(name))
    context = get_context(sample)
    context = get_sample_context()

    installer = context.installer
    return installer


@pytest.fixture(scope="session")
def session():
    return get_session("test")



from pytest  import fixture
from cdm.testing import get_context, get_installer

# returns a sample application
@fixture(scope="session")
def sample():
    context = get_context()


def test_simple_cql_schema():
    installer = get_installer("sample")
    schema = installer.cassandra_schema()
    assert len(schema) > 0



from gevent.monkey import patch_all
import logging
import os.path
import imp
import inspect
from base64 import b64encode
import urllib2

from gevent.pool import Pool
import progressbar

import pandas
from cdm.installer import Installer
from cdm.importer import Importer

class InstallerNotFound(Exception): pass


class Context(object):
    root = None
    session = None
    cache_dir = None
    dataset = None
    keyspace = None

    @property
    def installer(self):
        post_install = os.path.join(self.root, "install.py")

        logging.info("Loading installer {}".format(post_install))
        module = imp.load_source("Installer", post_install)
        members = inspect.getmembers(module)

        matching = [c for (name, c) in members if isinstance(c, type)
                    and c is not Installer
                    and issubclass(c, Installer)]

        if not matching:
            raise InstallerNotFound()

        installer = matching[0](self)
        self.feedback("Installer found")
        return installer

    def __init__(self, root, dataset, session, cache_dir):
        self.root = root
        self.dataset = dataset
        self.session = session
        # add trailing slash
        if not cache_dir.endswith("/"):
            cache_dir = cache_dir + "/"

        self.cache_dir = cache_dir
        if not os.path.exists(self.cache_dir):
            raise Exception("Cache dir does not exist")


    def download(self, url):
        """
        downloads a single file to the cache
        returns a file pointer
        auto caches download

        :param url:
        :param cache:
        :return:
        """
        encoded = b64encode(url)
        cache = self.cache_dir + encoded

        logging.info("Downloading {}".format(url))
        if not os.path.exists(cache):
            with open(cache, 'w') as fp:
                data = urllib2.urlopen(url)
                fp.write(data.read())
            logging.info("Download finished")
        else:
            self.feedback("Found in cache, skipping")

        fp = open(cache, 'r')
        return fp

    def open_local(self, name):
        tmp = os.path.join(self.root, name)
        return open(tmp, "r")

    def clean_cache(self):
        """
        removes all data from this cache directory
        :return: None
        """
        for file in os.listdir(self.cache_dir):
            logging.info(os.remove(self.cache_dir + file))


    def feedback(self, msg):
        logging.info(msg)

    def save_dataframe_to_cassandra(self, dataframe, table,
                                    transformation=None):
        """
        :param session:
        :type session: cassandra.cluster.Session
        :param dataframe:
        :type dataframe: pandas.DataFrame
        :param table:
        :return:
        """
        logging.info("Saving dataframe to %s", table)
        importer = Importer(self.session, dataframe, transformation)
        importer.load(table)


    def prepare(self, table, fields):
        field_list = ",".join(fields)
        placeholders = ",".join(["?"] * len(fields))
        statement = "INSERT INTO {} ({}) VALUES ({})".format(table, field_list, placeholders)
        logging.info("preparing %s", statement)
        return self.session.prepare(statement)



"""
notebook extension for all magic we need for tutorials
"""
from IPython.core.magic import Magics, magics_class, cell_magic, line_magic, needs_local_scope


@magics_class
class CDMMagic(Magics):
    pass



def load_ipython_extension(ipython):
    # set up the session and the spark context
    ipython.register_magics(CDMMagic)




import logging
from gevent.pool import Pool
from progressbar import ProgressBar

class Importer(object):
    session = None
    dataframe = None
    transformation = None

    def __init__(self, session, dataframe, transformation=None):
        """
        custom transformations accept a single row from a dataframe
        and should return a dictionary

        :param session:
        :param dataframe:
        :param transformation:
        :return:
        """
        self.session = session
        self.dataframe = dataframe
        self.transformation = transformation

    def transform(self, row):
        if self.transformation:
            return self.transformation(row)
        return dict(row)

    def iter(self):
        tmp = self.dataframe.iterrows()
        for i, row in tmp:
            yield self.transform(row)

    def get_insert(self, row_as_dict, table):
        fields = row_as_dict.keys()
        field_list = ",".join(fields)
        placeholders = ",".join(["?"] * len(fields))
        statement = "INSERT INTO {} ({}) VALUES ({})".format(table, field_list, placeholders)
        return (statement, row_as_dict.values())



    def load(self, table):
        cache = {}

        def save(row):
            (query, values) = self.get_insert(row, table)
            try:
                prepared = cache[query]
            except:
                prepared = self.session.prepare(query)
                cache[query] = prepared
            bound = prepared.bind(values)
            self.session.execute(bound)

        pool = Pool(100)
        i = 0
        print "Loading {}".format(table)
        with ProgressBar(max_value=len(self.dataframe)) as p:
            for _ in pool.imap_unordered(save, self.iter()):
                i += 1
                if i % 10 == 0:
                    p.update(i)









from cassandra.cqlengine.models import Model
from cassandra.cqlengine.columns import *

class Source(Model):
    name = Text(primary_key=True)
    url = Text()

class Dataset(Model):
    source = Text(primary_key=True)
    name = Text(primary_key=True)
    hash_installed = Text()



import os
from pytest import fixture

from cdm.context import Context
from cassandra.cluster import Cluster

from cdm.installer import Installer
from cdm.testing import get_context, get_sample_context

session = None

@fixture
def context():
    global session
    if not session:
        cluster = Cluster()
        session = cluster.connect()

        # create the keyspace
        ks = 'test_cdm'
        if 'test_cdm' not in cluster.metadata.keyspaces:
            q = "create KEYSPACE {} WITH \
                    replication = {{'class': \
                    'SimpleStrategy', \
                    'replication_factor': 1}}".format(ks)
            session.execute(q)
        if not os.path.exists("./cache"):
            os.mkdir("./cache")
        session.set_keyspace(ks)

    return Context(root=os.getcwd(),
                   dataset="test",
                   session=session,
                   cache_dir="./cache",
                   )



def test_download(context):
    context.clean_cache()
    fp = context.download("http://rustyrazorblade.com/pages/about-me.html")


def test_load_installer():
    sample = os.path.join(os.getcwd(), "tests/sample")
    context = get_context(sample)
    context = get_sample_context()

    installer = context.installer
    assert isinstance(installer, Installer)



"""cdm - Cassandra Data Manager

Usage:
    cdm list [<term>]
    cdm show <dataset>
    cdm install [--host=<host>] [--keyspace=<keyspace>] <dataset>
    cdm update
    cdm new <dataset>

Options:
    --debug

Examples:
    cdm install demo

"""

# for the future - cdm web & cdm stream

import sys

from subprocess import Popen

sys.path.append("")
import os
import os.path
import cdm
from ConfigParser import SafeConfigParser
import logging
from string import Template


# 3rd party

from docopt import docopt
from cdm.util import *

# allow for schema management fun
os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = "1"

# set up logging to console
console = logging.StreamHandler()
# console.setLevel(logging.INFO)
# set a format which is simpler for console use
# formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# console.setFormatter(formatter)
# add the handler to the root logger
logging_path = os.path.expanduser("~/cdm.log")
fh = logging.FileHandler(logging_path)
fh.setLevel(logging.INFO)

root_logger = logging.getLogger('')
root_logger.handlers = []
root_logger.addHandler(fh)
root_logger.setLevel(logging.INFO)


def main():
    arguments = docopt(__doc__)

    try:
        os.mkdir(CDM_CACHE)
    except OSError:
        pass


    if arguments["list"]:
        return list_datasets(arguments["<term>"])

    if arguments["update"]:
        return update_datasets()

    if arguments["install"]:
        tmp = arguments["<dataset>"].split("==")

        if len(tmp) == 1:
            tmp.append("master")

        # get defaults
        cqlshrc = os.path.expanduser("~/.cassandra/cqlshrc")
        if os.path.exists(cqlshrc):
            parser = SafeConfigParser()
            parser.read(cqlshrc)
            default_host = parser.get("connection", "hostname")
            default_port = parser.get("connection", "port")
        else:
            default_host = "localhost"

        host = arguments.get("--host") or default_host
        return install(tmp[0], tmp[1],
                       install_cassandra=True,# not arguments['--nocassandra'],
                       install_graph=False, # arguments['--graph'],
                       install_search=False, #arguments['--search'],
                       host=host)



    if arguments["show"]:
        return show_dataset_details(arguments["<dataset>"])
    #
    # if arguments["tutorials"]:
    #     # move into the tutorial directory
    #     # sys.argv["SPARK_DIRECTORY"] = ""
    #     if arguments["<dataset>"] == ".":
    #         os.chdir("tutorials")
    #     else:
    #         dest = os.path.join(CDM_CACHE,
    #                             "tutorials",
    #                             arguments["<dataset>"])
    #         os.chdir(dest)
    #
    #     if sys.platform.startswith('win'):
    #         p = Popen(["jupyter-notebook"], shell=True)
    #         # Don't raise KeyboardInterrupt in the parent process.
    #         # Set this after spawning, to avoid subprocess inheriting handler.
    #         import signal
    #         signal.signal(signal.SIGINT, signal.SIG_IGN)
    #         p.wait()
    #         sys.exit(p.returncode)
    #     else:
    #         os.execvp("jupyter-notebook", ['notebook'])

    if arguments["new"]:
        name = arguments["<dataset>"]

        logging.info("Creating %s directory", name)
        os.mkdir(name)
        os.chdir(name)
        logging.info("Creating tutorials")
        os.mkdir("tutorials")

        # set up the install template
        logging.info("Creating installer skeleton")
        tmp = os.path.dirname(cdm.__file__)
        tmp = os.path.join(tmp, os.path.pardir)
        root = os.path.abspath(tmp)
        skel = os.path.join(root, "skel", "install.py.template")
        skel = Template(open(skel).read())
        result = skel.substitute(name="MyInstaller")

        logging.info("Writing installer file")
        with open("install.py", 'w') as fp:
            fp.write(result)


    print "Done"





if __name__ == "__main__":
    main()



import os
import os.path
import subprocess
import urllib2
import logging
import yaml
from git import Repo
from importlib import import_module
from shutil import copytree, rmtree
from cassandra.cqlengine.connection import setup, get_session as get_db_session

# try to import the dse session
# if it works, monkey patch cqlengine to use dse session

from cdm.util import *
from cdm.context import Context
from cdm.installer import Installer
import sys
from distutils.spawn import find_executable
import yaml
DATASETS_URL = "https://raw.githubusercontent.com/cassandra-data-manager/cdm/master/datasets.yaml"

CDM_CACHE = os.getenv("CDM_CACHE", os.path.expanduser("~/.cdm/"))
CDM_PACKAGE_FILE = CDM_CACHE + "datasets.yaml"

logger = logging.getLogger(__name__)

def list_datasets(search):
    print "Available datasets:"
    data = open_datasets()

    #TODO rank results
    found = 0
    for name, details in data.iteritems():
        if search and search not in name and search not in details['description']:
            continue

        found += 1
        print "{:20} {}".format(name, details['description'])

    if found == 0:
        logger.info("No datasets found")


def open_datasets():
    with open(CDM_PACKAGE_FILE, 'r') as fp:
        data = yaml.load(fp)
    return data


def update_datasets():
    print "Updating datasets"
    fp = urllib2.urlopen(DATASETS_URL)
    data = fp.read()
    print data

    with open(CDM_PACKAGE_FILE, 'w') as d:
        d.write(data)

def normalize_dataset_name(dataset):
    return dataset.replace("-", "_")

def get_session(dataset, graph=False, host="localhost"):
    keyspace = normalize_dataset_name(dataset)
    if graph:
        try:
            from dse.cluster import Cluster
            from cassandra.cqlengine import connection
            connection.Cluster = Cluster
            print "Graph support enabled"
        except:
            pass

    session = connect(host=host, keyspace=keyspace)

    print "Creating keyspace"
    cql = "drop keyspace if exists {}".format(keyspace)
    session.execute(cql)

    cql = "create KEYSPACE {} WITH replication = {{'class': 'SimpleStrategy', 'replication_factor': 1}}".format(keyspace)
    session.execute(cql)

    if graph:
        graph_keyspace = keyspace + "_graph"


        try:
            session.execute_graph("system.dropGraph('{}')".format(graph_keyspace))
        except:
            pass

        session.execute_graph("system.createGraph('{}').build()".format(graph_keyspace))
        session.default_graph_options.graph_name = graph_keyspace
        print "Created graph keyspace {}".format(graph_keyspace)


    session.set_keyspace(keyspace)
    return session

def install(dataset,
            version="master",
            install_cassandra=True,
            install_graph=False,
            install_search=False,
            host=None):

    # check for cqlsh
    # if (install_graph or install_search) and not find_executable("dsetool"):
    #     print "DSE features requested but DSE bin not in PATH"
    #     sys.exit(1)

    if install_graph:
        try:
            import dse
        except:
            print "--graph supplied but python dse module not installed."
            sys.exit(1)

    # check for dse module

    if dataset == ".":
        path = "."
        dataset = "test"
        cache_dir = ".cdmcache"
    else:
        y = open_datasets()
        repo = download_dataset(dataset, y[dataset]['url'])

        # returns the git repo
        # we should be on master now
        # do I need a specific version?
        repo.git.checkout(version)
        path = local_dataset_path(dataset)
        # this should be keyspace
        cache_dir = CDM_CACHE + dataset + "_cache"
        sys.path.append(path) # so imports work

        tutorials_root_dir = os.path.join(CDM_CACHE, "tutorials")

        if not os.path.exists(tutorials_root_dir):
            logger.info("Creating tutorials directory")
            os.mkdir(tutorials_root_dir)

        logger.info("Setting up tutorials")
        dest = os.path.join(tutorials_root_dir, dataset)

        if os.path.exists(dest):
            logger.info("Clearing out old tutorial directory")
            rmtree(dest)

        copytree(os.path.join(path, "tutorials"), dest)

    # install requirements.  need to come back to this.
    # req_file = os.path.join(path, "requirements.txt")
    # if os.path.exists(req_file):
    #     print "Installing requirements"
    #     subprocess.check_call(["pip", "install", "-r", req_file])

    print "Connecting"

    session = get_session(dataset, install_graph, host=host)
    # load the schema

    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)

    context = Context(root=path,
                      dataset=dataset,
                      session=session,
                      cache_dir=cache_dir)


    # move to context
    installer = context.installer
    installer._cassandra = install_cassandra
    installer._search = install_search
    installer._graph = install_graph
    installer._install()



def local_dataset_path(dataset_name):
    return CDM_CACHE + dataset_name


def show_dataset_details(dataset_name):
    print "{}".format(dataset_name)

def download_dataset(dataset_name, dataset_url):
    local_git = local_dataset_path(dataset_name)
    if not os.path.exists(local_git):
        repo = Repo.clone_from(dataset_url, local_git)
        print "Downloaded"
    else:
        print "Repo exists, pulling latest"
        repo = Repo(local_git)
        git = repo.git
        git.checkout("master")
        repo.remotes[0].pull()
    return repo


# returns a new session
def connect(host="localhost", port=9042, keyspace=None, graph=False):
    setup([host], keyspace)
    return get_db_session()

def create_keyspace():
    # TODO ask for strategy and RF
    print "Creating keyspace (SimpleStrategy)"
    print "Replication factor 1"


def prompt(p, default):
    command = "{} [{}]>".format(p, default)
    result = raw_input(command).strip()
    return result if result else default


def configure_config(fp):
    cassandra_host = prompt("Cassandra host?", "localhost")
    spark_host = prompt("Spark host?", "localhost")

    # write config out
    data = {"cassandra_host": cassandra_host,
            "spark_host": spark_host }

    fp.write(yaml.dump(data, default_flow_style=False))




from distutils.core import setup
from setuptools import find_packages

setup(
        name='cassandra-dataset-manager',
        version='0.1',
        packages=find_packages(),
        url='https://github.com/riptano/cdm',
        license='License :: OSI Approved :: Apache License',
        author='Jon Haddad',
        author_email='jon@jonhaddad.com',
        description='Utility for quickly setting up sample datasets for Cassandra and DataStax Enterprise',
        entry_points={
            'console_scripts': [
                'cdm=cdm.cli:main'
            ]
        },
        install_requires=[
            "cassandra-driver>=3.1.0a2", "docopt>=0.6.1",
            "gitpython", "PyYAML", "pandas>=0.17.1",
            "colorama", "firehawk>=0.3",
            "fake-factory", "requests>=2.9.1",
            "gevent>=1.1.1", "progressbar2>=3.6.2"
        ]

)



from cdm.importer import Importer
from pandas import DataFrame
from pytest import fixture
from cdm.importer import Importer
from cdm.testing import session

@fixture
def df():
    data = [{'name': "Jon", 'hats': 1,
             'nickname':"Beast"},
            {'name': "Luke", 'hats': 50,
             'nickname':"Loopsie"}]

    return DataFrame(data)

@fixture
def importer(df, session):
    return Importer(session, df)


def test_simple_transform(session, df):
    i = Importer(session, df)
    row = i.iter().next()
    assert isinstance(row, dict)
    assert row['name'] == "Jon"

def test_custom_transform(session, df):
    def t(row):
        row['name'] = row['name'].upper()
        return dict(row)

    i = Importer(session, df, transformation=t)
    row = i.iter().next()
    assert isinstance(row, dict)
    assert row['name'] == "JON"


def test_insert_statement(importer):
    row = importer.iter().next()
    stmt, values = importer.get_insert(row, "tab")
    assert stmt.startswith("INSERT INTO tab")


def test_load(importer):
    queries = ["DROP TABLE IF EXISTS test_importer",
               "CREATE TABLE test_importer (name text primary key, hats int, nickname text)"]

    for q in queries:
        importer.session.execute(q)

    importer.load("test_importer")



import subprocess
import logging
import os
from abc import ABCMeta, abstractmethod
from collections import namedtuple
import requests
from cassandra.cqlengine.management import sync_table
from cassandra.cqlengine.models import ModelMetaClass

logger = logging.getLogger(__name__)

class SimpleCQLSchema(object):
    def cassandra_schema(self):
        data = open(self.schema, "r").read().split(";")
        return filter(lambda y: len(y), map(lambda x: x.strip(), data))

AutoGenerateSolrResources = namedtuple("AutoGenerateSolrResources", ["table"])
ExplicitSolrSchema = namedtuple("ExplicitSolrSchema", ["table"])

class Installer(object):
    __metaclass__ = ABCMeta

    context = None

    _search = False
    _graph = False
    _cassandra = True
    _cassandra_schema = True


    def __init__(self, context):
        # do not override
        self.context = context

    def post_init(self):
        # will get called after init
        logger.info("Post init, nothing to do.")

    def _install(self):
        self.post_init()
        logger.info("post_init() complete")

        if self._cassandra_schema:
            self.install_cassandra_schema()
        if self._cassandra:
            self.install_cassandra()
        if self._search:
            self.install_search_schema()
        if self._graph:
            self.install_graph_schema()
            self.install_graph()

        # set up tutorials

        logger.info("Done with install.")

    def install_cassandra_schema(self):
        # do not override
        logger.info("Applying schema {}".format(self.schema))

        self.context.session.set_keyspace(self.keyspace)
        for table in self.cassandra_schema():
            if isinstance(table, ModelMetaClass):
                sync_table(table)
            else:
                self.context.session.execute(table)


    def install_graph_schema(self):
        from firehawk.ddl import ParsedCommand
        for statement in self.graph_schema():
            logging.info("Schema %s", statement)
            if isinstance(statement, ParsedCommand):
                statement = str(statement)
            self.context.session.execute_graph(statement)


    def install_search_schema(self):
        logging.info("Setting up search schema")
        tables = self.search_schema()
        keyspace = self.keyspace
        host = "localhost"
        for table in tables:
            if isinstance(table, AutoGenerateSolrResources):
                url = "http://{}:8983/solr/admin/cores?action=CREATE&name={}.{}&generateResources=true&reindex=true".format(host, keyspace, table.table)
                logging.info(url)
                logging.info(requests.get(url))


    @abstractmethod
    def cassandra_schema(self):
        raise NotImplementedError("Cassandra schema required.")

    # @abstractmethod
    def graph_schema(self):
        raise NotImplementedError("Graph schema required.")

    # @abstractmethod
    def search_schema(self):
        # should return a list of SearchSchema classes
        raise NotImplementedError("Search schema required.")

    @abstractmethod
    def install_cassandra(self):
        raise NotImplementedError("Cassandra data required")


    # @abstractmethod
    def install_graph(self):
        logger.info("Graph requested but not implemented")
        raise NotImplementedError()

    @property
    def schema(self):
        return os.path.join(self.context.root, "schema.cql")

    @property
    def keyspace(self):
        return self.context.session.keyspace

Installer.register(SimpleCQLSchema)
